#!/usr/bin/env python3

# This program assembles source assembly code into a hex string.
# The hex string may include spaces and newlines for readability,
# these should be ignored by the simulator when it reads the corresponding file.

import os
import sys
import re
import string
import argparse

line=0 # global variable to make error reporting easier
current_address=0# idem
labels={} # global because shared between the two passe

def error(e):
    raise BaseException("Error at line " + str(line) + " : " + e)



# All the asm_xxxx functions are helper functions that parse an operand and convert it into its hex encoding.
# Feel free to add more

def asm_reg(s):
    "converts the string s into its encoding"
    if s=="0":
      return "0"
    if s=="1":
      return "1"  
    if s[0]!='r':
        error("invalid register: " + s)
    number = s[1:]
    if len(number)!=1:
        error("invalid register: " + s)
    if number not in "0123456789ABCDEFabcdef":
        error("invalid register: " + s)
    return number



def asm_offset_signed(s):
    "converts the string s into its encoding"
    # Is it a label or a constant? 
    if (s[0]>='0' and s[0]<='9') or s[0]=='-' or s[0]=='+': # TODO what it takes to catch hexa here
        val=int(s) # TODO  catch exception here
        # The following is not very elegant but easy to trust
        if val>=-128 and val<= 127:
            s  = hex(256+val)
            ss = s[-2:] # get rid of the 0x and possibly of the MSB hex digit
            return ss[::-1] # this is how you reverse a string in python. Needed here for little-endianness 
        else:
             error("offset out of range in   j " + s)
    else:
        error("Fixme! labels currently unsupported")



def asm_pass(iteration, s_file):
    global line
    global labels
    global current_address
    code =[] # array of strings, one entry per instruction
    print("\n PASS " + str(iteration))
    current_address = 1024
    source = open(s_file)
    for source_line in source:
        instruction_encoding="" 
        print("processing line: " + source_line[0:-1]) # just to get rid of the final newline

        # if there is a comment, get rid of it
        index = source_line.find(";")
        if index !=-1:
            source_line = source_line[:index]

        # split the non-comment part of the line into tokens (thanks Stack Overflow) 
        tokens = re.findall('[\S]+', source_line) # \S means: any non-whitespace
        print(tokens) # to debug

        # if there is a label, consume it
        if tokens:
            token=tokens[0]
            if token[-1] == ":": # last character
                label = token[0: -1] # all the characters except last one
                print(" ...found label declaration: " + label + "  at address " + str(current_address))
                if  label in labels:
                    error("error: this label was already defined, at address " + str(labels[label]))
                labels[label] = current_address
                tokens = tokens[1:]

        # now all that remains should be an instruction... or nothing
        if tokens:
            opcode = tokens[0]
            token_count = len(tokens)
            if opcode == "j" and token_count==2:
                    instruction_encoding = "0" + asm_offset_signed(tokens[1])
            if opcode == "add" and token_count==3:
                    instruction_encoding = "3" + asm_reg(tokens[1]) + asm_reg(tokens[2])
            # Now a lot of constructive copy-paste, for instance
            if opcode == "sub" and token_count==3:
                    instruction_encoding = "4" + asm_reg(tokens[1]) + asm_reg(tokens[2])
         #begin sabote
         #end sabote
                    
            # If the line wasn't assembled:
            if instruction_encoding=="":
                error("don't know what to do with: " + source_line)
            else:
                instr_size = len(instruction_encoding)
                # Debug output
                print(" ... @ " + str(current_address) + "  :  " + instruction_encoding )
                current_address += instr_size
                code.append(instruction_encoding)

                
        line += 1
    source.close()
    return code




#/* main */
if __name__ == '__main__':

    argparser = argparse.ArgumentParser(description='This is the assembler for the ASR2017 processor @ ENS-Lyon')
    argparser.add_argument('filename', help='name of the source file.  "python asm.py toto.s" assembles toto.s into toto.obj')

    options=argparser.parse_args()
    filename = options.filename
    basefilename, extension = os.path.splitext(filename)
    obj_file = basefilename+".obj"
    code = asm_pass(1, filename) # first pass essentially builds the labels

     # second pass is for good, but is disabled now:
     # re-enable it to support labels
     # code = asm_pass(2, filename)

    # statistics
    print("Average instruction size is " + str(1.0*(current_address-1024)/len(code)) + " nibbles" ) 
    
    outfile = open(obj_file, "w")
    for instr in code:
        outfile.write(instr)
        outfile.write("\n")


    outfile.close()
