#include "processor.h"

Processor::Processor(Memory* m): m(m)
{
	pc=1024; // start the program at address 0
	for (int i=0; i<16; i++)
		r[i]=0;
	r[1]=1; // and it should never change again
	flag=0;
}

Processor::~Processor()
{}



void Processor::von_Neuman_step(bool debug) {
	// various variables for convenience
	uint32_t old_pc;


	if (debug)
		{  // show state before the instruction
			std::cout << "Before instruction: pc=" << std::hex << std::setw(8) << std::setfill('0') << pc  <<
				" pointing to ";
			for (int i=0; i<16; i++)  				std::cout << std::hex << std::setw(1) << m->read(pc+i);
			std::cout <<		"..." << 	std::endl;
		}

	old_pc = pc; // mostly for debug

	
	// Fetch first nibble of the instruction	
	uint32_t instr = m->read(pc);  	pc = pc+1; // yes I know I could write  m->read(pc++)  but I won't.

	// Execute
	switch(instr)
		{
		case 0x0:  // j (reative jump)
			{
				// read two more nibbles
				int32_t delta; 
				delta = (m->read(pc))<<24; // put it on top of a 32-bit word
				pc++;
				delta += (m->read(pc))<<28; //
				pc++;
				delta = delta >> 24; // shift it back to the lower byte: this performs the sign extension
				//std::cout << "delta=" << delta << std::endl;
				pc = old_pc + delta;
				break;
			}
		case 0x3: // addition, I don't manage carries yet: TODO
			{
				uint32_t x, y;
				x = m->read(pc);  pc++;
				y = m->read(pc);  pc++;
				if(x!=0 && x!=1) {
					r[x] = r[x] + r[y];
				}
				break; // don't forget these breaks. That's C for you.
			}
			// begin sabote
			//end sabote
		default:
		std::cerr << "Illegal instruction: " << std::hex << instr << " at PC=" << pc << std::endl;
		// and that's it, we proceed with execution
		break;
		
		}


	if (debug)
		{ 
			std::cout << "After instruction,  pc="	<< std::hex << std::setw(8) << std::setfill('0') << pc;
			for (int i=0; i<16; i++)
				std::cout << ((i&3)==0?  "\n" : "") << " r" << std::hex << i <<" = "
									<< std::hex << std::setw(8) << std::setfill('0') << r[i] << " ";
		}
	std::cout <<	std::endl;
}


